@@ -220,13 +220,13 @@ main(int argc, char *argv[])
 	const char *repodir = NULL;
 	int i, ret = 0;
 
-	if (pledge("stdio rpath", NULL) == -1)
-		err(1, "pledge");
-
 	setlocale(LC_CTYPE, "");
 
 	git_libgit2_init();
 
+	if (pledge("stdio rpath", NULL) == -1)
+		err(1, "pledge");
+
 	for (i = 1; i < argc; i++) {
 		if (argv[i][0] == '-') {
 			if (argv[i][1] != 'b' || i + 1 >= argc)
@@ -273,6 +273,10 @@ main(int argc, char *argv[])
 
 		writelog(stdout);
 	}
+	if (!repodir) {
+		fprintf(stderr, "%s [-b baseprefix] [repodir...]\n", argv[0]);
+		return 1;
+	}
 
 	/* cleanup */
 	git_repository_free(repo);
