@@ -58,7 +58,8 @@ static char *name = "";
 static char *strippedname = "";
 static char description[255];
 static char cloneurl[1024];
-static int haslicense, hasreadme, hassubmodules;
+static int haslicense, hasreadme, hassubmodules, hastagsgph;
+static long long nlogcommits = -1; /* < 0 indicates not used */
 
 /* cache */
 static git_oid lastoid;
@@ -75,29 +76,33 @@ static const char *cachefile;
 int
 utf8pad(char *buf, size_t bufsiz, const char *s, size_t len, int pad)
 {
-	wchar_t w;
+	wchar_t wc;
 	size_t col = 0, i, slen, siz = 0;
-	int rl, wc;
+	int rl, w;
 
 	if (!len)
 		return -1;
 
 	slen = strlen(s);
-	for (i = 0; i < slen && col < len + 1; i += rl) {
-		if ((rl = mbtowc(&w, &s[i], slen - i < 4 ? slen - i : 4)) <= 0)
+	for (i = 0; i < slen; i += rl) {
+		if ((rl = mbtowc(&wc, &s[i], slen - i < 4 ? slen - i : 4)) <= 0)
 			break;
-		if ((wc = wcwidth(w)) == -1)
-			wc = 1;
-		col += wc;
-		if (col >= len && s[i + rl]) {
+		if ((w = wcwidth(wc)) == -1)
+			continue;
+		if (col + w > len || (col + w == len && s[i + rl])) {
 			if (siz + 4 >= bufsiz)
 				return -1;
-			memcpy(&buf[siz], "\xe2\x80\xa6", 4);
+			memcpy(&buf[siz], "\xe2\x80\xa6", 3);
+			siz += 3;
+			if (col + w == len && w > 1)
+				buf[siz++] = pad;
+			buf[siz] = '\0';
 			return 0;
 		}
 		if (siz + rl + 1 >= bufsiz)
 			return -1;
 		memcpy(&buf[siz], &s[i], rl);
+		col += w;
 		siz += rl;
 		buf[siz] = '\0';
 	}
@@ -130,7 +135,7 @@ deltainfo_free(struct deltainfo *di)
 	if (!di)
 		return;
 	git_patch_free(di->patch);
-	di->patch = NULL;
+	memset(di, 0, sizeof(*di));
 	free(di);
 }
 
@@ -138,6 +143,7 @@ int
 commitinfo_getstats(struct commitinfo *ci)
 {
 	struct deltainfo *di;
+	git_diff_options opts;
 	const git_diff_delta *delta;
 	const git_diff_hunk *hunk;
 	const git_diff_line *line;
@@ -145,6 +151,20 @@ commitinfo_getstats(struct commitinfo *ci)
 	size_t ndeltas, nhunks, nhunklines;
 	size_t i, j, k;
 
+	if (git_tree_lookup(&(ci->commit_tree), repo, git_commit_tree_id(ci->commit)))
+		goto err;
+	if (!git_commit_parent(&(ci->parent), ci->commit, 0)) {
+		if (git_tree_lookup(&(ci->parent_tree), repo, git_commit_tree_id(ci->parent))) {
+			ci->parent = NULL;
+			ci->parent_tree = NULL;
+		}
+	}
+
+	git_diff_init_options(&opts, GIT_DIFF_OPTIONS_VERSION);
+	opts.flags |= GIT_DIFF_DISABLE_PATHSPEC_MATCH;
+	if (git_diff_tree_to_tree(&(ci->diff), repo, ci->parent_tree, ci->commit_tree, &opts))
+		goto err;
+
 	ndeltas = git_diff_num_deltas(ci->diff);
 	if (ndeltas && !(ci->deltas = calloc(ndeltas, sizeof(struct deltainfo *))))
 		err(1, "calloc");
@@ -186,6 +206,14 @@ commitinfo_getstats(struct commitinfo *ci)
 	return 0;
 
 err:
+	git_diff_free(ci->diff);
+	ci->diff = NULL;
+	git_tree_free(ci->commit_tree);
+	ci->commit_tree = NULL;
+	git_tree_free(ci->parent_tree);
+	ci->parent_tree = NULL;
+	git_commit_free(ci->parent);
+	ci->parent = NULL;
 	if (ci->deltas)
 		for (i = 0; i < ci->ndeltas; i++)
 			deltainfo_free(ci->deltas[i]);
@@ -210,12 +238,12 @@ commitinfo_free(struct commitinfo *ci)
 		for (i = 0; i < ci->ndeltas; i++)
 			deltainfo_free(ci->deltas[i]);
 	free(ci->deltas);
-	ci->deltas = NULL;
 	git_diff_free(ci->diff);
 	git_tree_free(ci->commit_tree);
 	git_tree_free(ci->parent_tree);
 	git_commit_free(ci->commit);
 	git_commit_free(ci->parent);
+	memset(ci, 0, sizeof(*ci));
 	free(ci);
 }
 
@@ -223,7 +251,6 @@ struct commitinfo *
 commitinfo_getbyoid(const git_oid *id)
 {
 	struct commitinfo *ci;
-	git_diff_options opts;
 
 	if (!(ci = calloc(1, sizeof(struct commitinfo))))
 		err(1, "calloc");
@@ -240,20 +267,6 @@ commitinfo_getbyoid(const git_oid *id)
 	ci->summary = git_commit_summary(ci->commit);
 	ci->msg = git_commit_message(ci->commit);
 
-	if (git_tree_lookup(&(ci->commit_tree), repo, git_commit_tree_id(ci->commit)))
-		goto err;
-	if (!git_commit_parent(&(ci->parent), ci->commit, 0)) {
-		if (git_tree_lookup(&(ci->parent_tree), repo, git_commit_tree_id(ci->parent))) {
-			ci->parent = NULL;
-			ci->parent_tree = NULL;
-		}
-	}
-
-	git_diff_init_options(&opts, GIT_DIFF_OPTIONS_VERSION);
-	opts.flags |= GIT_DIFF_DISABLE_PATHSPEC_MATCH;
-	if (git_diff_tree_to_tree(&(ci->diff), repo, ci->parent_tree, ci->commit_tree, &opts))
-		goto err;
-
 	return ci;
 
 err:
@@ -297,22 +310,17 @@ gphtextnl(FILE *fp, const char *s, size_t len)
 {
 	size_t i, n = 0;
 
-	for (i = 0; *s && i < len; i++) {
-		if (s[i] == '\n')
-			n = 0;
-
+	for (i = 0; s[i] && i < len; i++) {
 		/* escape with 't' at the start of a line */
-		if (!n && (s[i] == 't' || s[i] == '[')) {
+		if (!n && (s[i] == 't' || s[i] == '['))
 			fputc('t', fp);
-			n = 1;
-		}
 
 		switch (s[i]) {
+		case '\t': fputs("        ", fp);
 		case '\r': break;
-		case '\t': fputs("        ", fp); break;
 		default: fputc(s[i], fp);
 		}
-		n++;
+		n = (s[i] != '\n');
 	}
 }
 
@@ -323,8 +331,8 @@ gphtext(FILE *fp, const char *s, size_t len)
 {
 	size_t i;
 
-	for (i = 0; *s && i < len; i++) {
-		switch (s[i]) {
+	for (i = 0; *s && i < len; s++, i++) {
+		switch (*s) {
 		case '\r': /* ignore CR */
 		case '\n': /* ignore LF */
 			break;
@@ -332,7 +340,7 @@ gphtext(FILE *fp, const char *s, size_t len)
 			fputs("        ", fp);
 			break;
 		default:
-			fputc(s[i], fp);
+			fputc(*s, fp);
 			break;
 		}
 	}
@@ -344,8 +352,8 @@ gphlink(FILE *fp, const char *s, size_t len)
 {
 	size_t i;
 
-	for (i = 0; *s && i < len; i++) {
-		switch (s[i]) {
+	for (i = 0; *s && i < len; s++, i++) {
+		switch (*s) {
 		case '\r': /* ignore CR */
 		case '\n': /* ignore LF */
 			break;
@@ -356,7 +364,7 @@ gphlink(FILE *fp, const char *s, size_t len)
 			fputs("\\|", fp);
 			break;
 		default:
-			fputc(s[i], fp);
+			fputc(*s, fp);
 			break;
 		}
 	}
@@ -451,6 +459,8 @@ writeheader(FILE *fp, const char *title)
 	fprintf(fp, "[1|Log|%s/log.gph|server|port]\n", relpath);
 	fprintf(fp, "[1|Files|%s/files.gph|server|port]\n", relpath);
 	fprintf(fp, "[1|Refs|%s/refs.gph|server|port]\n", relpath);
+	if (hastagsgph)
+		fprintf(fp, "[1|Tags|%s/tags.gph|server|port]\n", relpath);
 	if (hassubmodules)
 		fprintf(fp, "[1|Submodules|%s/file/.gitmodules.gph|server|port]\n", relpath);
 	if (hasreadme)
@@ -652,8 +662,7 @@ writelogline(FILE *fp, struct commitinfo *ci)
 	fputs("  ", fp);
 	utf8pad(buf, sizeof(buf), ci->author ? ci->author->name : "", 25, ' ');
 	gphlink(fp, buf, strlen(buf));
-	fprintf(fp, "|%s/commit/%s.gph", relpath, ci->oid);
-	fputs("|server|port]\n", fp);
+	fprintf(fp, "|%s/commit/%s.gph|server|port]\n", relpath, ci->oid);
 }
 
 int
@@ -662,7 +671,7 @@ writelog(FILE *fp, const git_oid *oid)
 	struct commitinfo *ci;
 	git_revwalk *w = NULL;
 	git_oid id;
-	char path[PATH_MAX];
+	char path[PATH_MAX], oidstr[GIT_OID_HEXSZ + 1];
 	FILE *fpfile;
 	int r;
 
@@ -674,20 +683,34 @@ writelog(FILE *fp, const git_oid *oid)
 	while (!git_revwalk_next(&id, w)) {
 		if (cachefile && !memcmp(&id, &lastoid, sizeof(id)))
 			break;
+
+		git_oid_tostr(oidstr, sizeof(oidstr), &id);
+		r = snprintf(path, sizeof(path), "commit/%s.gph", oidstr);
+		if (r == -1 || (size_t)r >= sizeof(path))
+			errx(1, "path truncated: 'commit/%s.gph'", oidstr);
+		r = access(path, F_OK);
+
+		/* optimization: if there are no log lines to write and
+		   the commit file already exists: skip the diffstat */
+		if (!nlogcommits && !r)
+			continue;
+
 		if (!(ci = commitinfo_getbyoid(&id)))
 			break;
 
-		writelogline(fp, ci);
+		if (nlogcommits < 0) {
+			writelogline(fp, ci);
+		} else if (nlogcommits > 0) {
+			writelogline(fp, ci);
+			nlogcommits--;
+		}
+
 		if (cachefile)
 			writelogline(wcachefp, ci);
 
-		r = snprintf(path, sizeof(path), "commit/%s.gph", ci->oid);
-		if (r == -1 || (size_t)r >= sizeof(path))
-			errx(1, "path truncated: 'commit/%s.gph'", ci->oid);
-
 		/* check if file exists if so skip it */
-		if (access(path, F_OK)) {
-			/* lookup stats: only required here */
+		if (r) {
+			/* lookup stats: only required here for gopher */
 			if (commitinfo_getstats(ci) == -1)
 				goto err;
 
@@ -1074,7 +1097,7 @@ err:
 void
 usage(char *argv0)
 {
-	fprintf(stderr, "%s [-b baseprefix] [-c cachefile] repodir\n", argv0);
+	fprintf(stderr, "%s [-b baseprefix] [-c cachefile] [-l commits] repodir\n", argv0);
 	exit(1);
 }
 
@@ -1091,9 +1114,6 @@ main(int argc, char *argv[])
 	size_t n;
 	int i, fd;
 
-	if (pledge("stdio rpath wpath cpath", NULL) == -1)
-		err(1, "pledge");
-
 	setlocale(LC_CTYPE, "");
 
 	for (i = 1; i < argc; i++) {
@@ -1106,9 +1126,20 @@ main(int argc, char *argv[])
 				usage(argv[0]);
 			relpath = argv[++i];
 		} else if (argv[i][1] == 'c') {
-			if (i + 1 >= argc)
+			if (nlogcommits > 0 || i + 1 >= argc)
 				usage(argv[0]);
 			cachefile = argv[++i];
+		} else if (argv[i][1] == 'l') {
+			if (cachefile || i + 1 >= argc)
+				usage(argv[0]);
+			errno = 0;
+			nlogcommits = strtoll(argv[++i], &p, 10);
+			if (argv[i][0] == '\0' || *p != '\0' ||
+			    nlogcommits <= 0)
+				usage(argv[0]);
+			if (errno == ERANGE && (nlogcommits == LLONG_MAX ||
+			    nlogcommits == LLONG_MIN))
+				usage(argv[0]);
 		}
 	}
 	if (!repodir)
@@ -1119,6 +1150,14 @@ main(int argc, char *argv[])
 
 	git_libgit2_init();
 
+	if (cachefile) {
+		if (pledge("stdio rpath wpath cpath fattr", NULL) == -1)
+			err(1, "pledge");
+	} else {
+		if (pledge("stdio rpath wpath cpath", NULL) == -1)
+			err(1, "pledge");
+	}
+
 	if (git_repository_open_ext(&repo, repodir,
 		GIT_REPOSITORY_OPEN_NO_SEARCH, NULL) < 0) {
 		e = giterr_last();
@@ -1131,10 +1170,6 @@ main(int argc, char *argv[])
 		head = git_object_id(obj);
 	git_object_free(obj);
 
-	/* don't cache if there is no HEAD */
-	if (!head)
-		cachefile = NULL;
-
 	/* use directory name as name */
 	if ((name = strrchr(repodirabs, '/')))
 		name++;
@@ -1173,6 +1208,9 @@ main(int argc, char *argv[])
 		fclose(fpread);
 	}
 
+	/* check if tags.gph file exists */
+	hastagsgph = (access("tags.gph", F_OK) == 0);
+
 	/* check LICENSE */
 	haslicense = (!git_revparse_single(&obj, repo, "HEAD:LICENSE") &&
 		git_object_type(obj) == GIT_OBJ_BLOB);
@@ -1196,7 +1234,7 @@ main(int argc, char *argv[])
 	fprintf(fp, "%-50.50s  ", "Commit message");
 	fprintf(fp, "%s\n", "Author");
 
-	if (cachefile) {
+	if (cachefile && head) {
 		/* read from cache file (does not need to exist) */
 		if ((rcachefp = fopen(cachefile, "r"))) {
 			if (!fgets(lastoidstr, sizeof(lastoidstr), rcachefp))
@@ -1258,7 +1296,7 @@ main(int argc, char *argv[])
 	fclose(fp);
 
 	/* rename new cache file on success */
-	if (cachefile) {
+	if (cachefile && head) {
 		if (rename(tmppath, cachefile))
 			err(1, "rename: '%s' to '%s'", tmppath, cachefile);
 		umask((mask = umask(0)));
